# -*- coding: utf-8 -*-

""" GeoMcFly - Geographically localized netalinks created on the fly.

This module is for creation of metalinks from a list of mirrors with their preference
relative to clients' geographical location.

While it's primary target of use is for the Mandriva Linux distribution and
downloading of it's packages, it aims to be written in a generic and portable way
making it usable for other purposes as well.

Copyright © Per Øyvind Karlsen <peroyvind@mandriva.org> 2007

This is free software and may be redistributed under the terms of the GNU GPL.

"""

from os import access, getenv, stat, F_OK, R_OK
from metalink import Metalink
from GeoIP import open as GeoOpen
from GeoIP import GEOIP_STANDARD

class GeoMcFly():
	def __init__(self, listfile):
		self.version = 0.8.1
		self.datadir = '/var/lib/geomcfly'
		from os.path import join
		self.listfile = join(self.datadir,listfile)
		self.geoip = GeoOpen('/usr/share/GeoIP/GeoLiteCity.dat', GEOIP_STANDARD)
		self.metalink = Metalink()
		self.metalink.origin = 'Geolocalized metalink created on the fly by Geo McFly %s' % 0.9
		self.metalink.type = 'dynamic'
		self.mirrorlist = []
		self.lock='/var/run/geomcfly/lock'

	def loadlist(self, listurl, reqtype='', maxage=86400):
		"""
		Loads the object with mirrors and their geoip data from from a file. If the
		object file is missing or older than maxage it will be recreated.

		@param listurl: The url to download the mirror list from.
		@param reqtype: The type of mirrors to obtain. Can be 'distrib', 'updates'
		or 'iso'. (Mandriva specific)
		@param maxage: The max age of the object file in seconds. Default is
		864000 (one day).
		"""
		from time import time
		mirrorlist = []
		if access(self.listfile,R_OK) and ((time() - stat(self.listfile).st_mtime) < maxage):
			file = open(self.listfile, 'r')
			from pickle import load
			mirrorlist = load(file)
			file.close()
		else:
			mirrorlist = self.createlist(listurl, reqtype)

		if isinstance(mirrorlist, list):
			if isinstance(mirrorlist[0], dict):
				self.mirrorlist = mirrorlist
				return True
		return False

	def createlist(self, listurl, reqtype=''):
		"""
		Downloads the list of mirrors, adds geoip their geoip data and saves
		them to an object file. (Very Mandriva specific, needs to be adapted for
		other purposes.)
		
		@param listurl: The url to download the mirror list from.
		@param reqtype: The type of mirrors to obtain. Can be 'distrib', 'updates'
		or 'iso'. (Mandriva specific)
		"""
		
		from curl import Curl
		liste = Curl().get(listurl).split('\n')[:-1]

		mirrorlist = []
		for mirror in liste[:]:
			city, bandwidth, url, type, connections = None, None, None, None, ''
			for field in mirror.split(',')[:]:
				detail, value = field.split('=')
				if detail == 'city':
					city = value
				if detail == 'bw':
					bandwidth = detail
				if detail == 'url':
					url = value
				if detail == 'connections':
					connections = value
				if detail == 'type':
					type = value
			if reqtype != type:
				continue
			host = url.split('/')[2]
			mirrordata = self.geoip.record_by_name(host)
			mirrordata['host'] = host
			mirrordata['url'] = url
			mirrordata['preference'] = ''
			mirrordata['bandwidth'] = bandwidth
			mirrordata['connections'] = connections
			mirrorlist.append(mirrordata)

		# Simple lock in hope of avoiding possible locking issues
		from os import remove
		from time import time
		locked = False
		if access(self.lock,F_OK):
			# If lock is more than a minute old, consider it stale and remove it
			if (time() - stat(self.lock).st_mtime > 60):
				remove(self.lock)
			else:
				locked = True

		if not locked:
			lockfile = open(self.lock, 'w')
			lockfile.write('geomc update lock')
			lockfile.close()
			file = open(self.listfile, 'w')
			from pickle import dump
			dump(mirrorlist, file)
			file.close()
			remove(self.lock)

		return mirrorlist

	def calcpref(self, ipaddr):
		"""
		Calculates mirror preference relative to ip address.

		@param ipaddr: The ip address to calculate against.
		"""
		client = self.geoip.record_by_addr(ipaddr)
		proximity = {}
		for mirror in self.mirrorlist[:]:
			host = mirror['host']
			if proximity.has_key(host):
				continue
			diff = {}
			diff['longitude'] = client['longitude'] - mirror['longitude']
			diff['latitude'] = client['latitude'] - mirror['latitude']
			for key in diff.iterkeys():
				if diff[key] < 0:
					diff[key] = diff[key] * -1
			proximity[host] = 100 - ((diff['longitude']*180.0) + (diff['latitude']*90.0)) / (90.0+180.0)
		values = proximity.values()
		values.sort(cmp,reverse=True)

		for mirror in self.mirrorlist[:]:
			proxval = proximity[mirror['host']]
			mirror['preference'] = 100 - values.index(proxval)


	def addfiles(self,files):
		"""
		Adds files to the metalink.

		@param files: Array of strings with file names.
		"""
		skip = True
		for file in files[:]:
			# Metalink object starts with a file, so use it as a first and only create
			# new files for the rest.
			if skip:
				skip = False
			else:
				self.metalink.add_file()
			try:
				from os.path import basename
				self.metalink.scan_file(basename(file))
			except:
				None
			for mirror in self.mirrorlist[:]:
				url = '%s/%s' % (mirror['url'], file)
				self.metalink.add_url(url,location=mirror['country_code'],preference=mirror['preference'],conns=mirror['connections'])

	def getmetalink(self):
		"""
		Generates and returns the metalink.
		"""
		return self.metalink.generate()
